//
//  AppDelegate.swift
//  Encoder
//
//  Created by Phidgets on 2018-02-22.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = Phidget22Swift.Encoder()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!
    
    //GUI
    @IBOutlet weak var settingsBox: NSBox!
    @IBOutlet weak var extraSettingsBox: NSBox!
    @IBOutlet weak var ioModeBox: NSBox!
    @IBOutlet weak var dataBox: NSBox!
    @IBOutlet weak var enabledCheck: NSButton!
    @IBOutlet weak var cprText: NSTextField!
    @IBOutlet weak var dataIntervalSlider: NSSlider!
    @IBOutlet weak var minDataIntervalLabel: NSTextField!
    @IBOutlet weak var maxDataIntervalLabel: NSTextField!
    @IBOutlet weak var dataIntervalLabel: NSTextField!
    @IBOutlet weak var changeTriggerSlider: NSSlider!
    @IBOutlet weak var minChangeTriggerLabel: NSTextField!
    @IBOutlet weak var maxChangeTriggerLabel: NSTextField!
    @IBOutlet weak var changeTriggerLabel: NSTextField!
    @IBOutlet weak var ioModePopup: NSPopUpButton!
    @IBOutlet weak var velocityLabel: NSTextField!
    @IBOutlet weak var indexPositionLabel: NSTextField!
    @IBOutlet weak var positionLabel: NSTextField!
    @IBOutlet weak var timeChangeLabel: NSTextField!
    @IBOutlet weak var positionChangeLabel: NSTextField!
    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0
    
    //CPR Calculation
    var CPREnabled = false
    var CPRValue = 0.0
    
    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            let _ = ch.positionChange.addHandler(positionchange_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            ch.positionChange.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! Phidget22Swift.Encoder
            
            let deviceID = try attachedDevice.getDeviceID()
            let minChangeTrigger = try attachedDevice.getMinPositionChangeTrigger()
            let maxChangeTrigger = try attachedDevice.getMaxPositionChangeTrigger()
            let changeTrigger = try attachedDevice.getPositionChangeTrigger()
            let minDataInterval = try attachedDevice.getMinDataInterval()
            let maxDataInterval = try attachedDevice.getMaxDataInterval()
            let dataInterval = try attachedDevice.getDataInterval()
            var enabled:Bool?
            var ioMode:EncoderIOMode?
            var heightchange:CGFloat = 0.0
            
            if(deviceID == DeviceID.PN_1047 || deviceID == DeviceID.PN_ENC1000){
                enabled = try ch.getEnabled()
            }
            
            if(deviceID == DeviceID.PN_DCC1000 || deviceID == DeviceID.PN_ENC1000){
                ioMode = try attachedDevice.getIOMode()
            }
            
            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: sender)
                self.dataBox.isHidden = false
                self.settingsBox.isHidden = false
                self.extraSettingsBox.isHidden = false
                
                if(deviceID == DeviceID.PN_1047 || deviceID == DeviceID.PN_ENC1000){
                    self.enabledCheck.state = enabled! ? NSControl.StateValue.on : NSControl.StateValue.off
                }
                
                if(deviceID == DeviceID.PN_DCC1000 || deviceID == DeviceID.PN_ENC1000){
                    self.ioModeBox.isHidden = false
                    heightchange += self.ioModeBox.frame.size.height
                    self.ioModePopup.addItem(withTitle: "Push Pull")
                    self.ioModePopup.addItem(withTitle: "Line Driver 2.2k")
                    self.ioModePopup.addItem(withTitle: "Line Driver 10k")
                    self.ioModePopup.addItem(withTitle: "Open Collector 2.2k")
                    self.ioModePopup.addItem(withTitle: "Open Collector 10k")
                    self.ioModePopup.selectItem(at: ioMode == EncoderIOMode.pushPull ? 0 : ioMode == EncoderIOMode.lineDriver_2K2 ? 1 : ioMode == EncoderIOMode.lineDriver_10K ? 2 : ioMode == EncoderIOMode.openCollector_2K2 ? 3 : 4)
                }
                
                self.dataIntervalSlider.minValue = Double(minDataInterval)
                self.dataIntervalSlider.maxValue = Double(maxDataInterval)
                self.dataIntervalSlider.doubleValue = Double(dataInterval)
                
                self.minDataIntervalLabel.stringValue = String(format:"%d",minDataInterval)
                self.maxDataIntervalLabel.stringValue = String(format:"%d",maxDataInterval)
                self.dataIntervalLabel.stringValue = String(format:"%d ms",dataInterval)
                
                self.changeTriggerSlider.minValue = Double(minChangeTrigger)
                self.changeTriggerSlider.maxValue = Double(maxChangeTrigger)
                self.changeTriggerSlider.doubleValue = Double(changeTrigger)
                
                self.minChangeTriggerLabel.stringValue = String(format:"%d", minChangeTrigger)
                self.maxChangeTriggerLabel.stringValue = String(format:"%d", maxChangeTrigger)
                self.changeTriggerLabel.stringValue = String(format:"%d counts", changeTrigger)
                
                
                //Adjusting window height
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height + heightchange
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.dataBox.isHidden = true
            self.settingsBox.isHidden = true
            self.extraSettingsBox.isHidden = true
            self.ioModeBox.isHidden = true
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    func positionchange_handler(sender: Phidget22Swift.Encoder, data: (positionChange: Int, timeChange: Double, indexTriggered: Bool)){
        do{
            let position = try ch.getPosition()
            DispatchQueue.main.async {
                self.positionChangeLabel.stringValue = String(format:"%d counts", data.positionChange)
                self.timeChangeLabel.stringValue = String(format:"%.2f ms", data.timeChange)
                self.positionLabel.stringValue = String(format:"%lli counts", position)
            }
            
            if(data.indexTriggered){
                let indexPosition = try ch.getIndexPosition()
                DispatchQueue.main.async {
                    self.indexPositionLabel.stringValue = String(format:"%lli counts", indexPosition)
                }
            }
            
            if(CPREnabled){
                DispatchQueue.main.async {
                    self.velocityLabel.stringValue = String(format:"%.1f RPM", (Double(data.positionChange) * 60000.0) / (data.timeChange * self.CPRValue))
                }
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: GUI Control
    @IBAction func setEnabled(_ sender: Any) {
        do{
            let check = sender as! NSButton
            try ch.setEnabled(check.state == NSControl.StateValue.on ? true : false)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setDataInterval(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setDataInterval(UInt32(slider.intValue))
            dataIntervalLabel.stringValue = String(format:"%d ms", UInt32(slider.intValue))
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setChangeTrigger(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setPositionChangeTrigger(UInt32(slider.intValue))
            changeTriggerLabel.stringValue = String(format:"%d counts", UInt32(slider.intValue))
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setCPR(_ sender: Any) {
        if let cprvalue = Double(cprText.stringValue){
            CPREnabled = true
            CPRValue = cprvalue
        }
        else{
            cprText.stringValue = "Invalid"
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }
    
}

